"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

import logging
from _socket import SOL_SOCKET, SOL_TCP, SO_KEEPALIVE, TCP_KEEPIDLE, TCP_KEEPINTVL, TCP_KEEPCNT
from _ssl import PROTOCOL_TLSv1
from threading import Lock

import gevent
import socks
from gevent import GreenletExit

from pysoltcp import PY2

from gevent.queue import Queue
from pysolbase.SolBase import SolBase
from pysolmeters.AtomicInt import AtomicIntSafe

from pysoltcp.tcpbase.ProtocolParserTextDelimited import ProtocolParserTextDelimited
from pysoltcp.tcpbase.TcpSocketManager import TcpSocketManager
from pysoltcp.tcpclient.TcpClientConfig import TcpClientConfig

if PY2:
    # noinspection PyProtectedMember
    from gevent._sslgte279 import SSLSocket
else:
    # noinspection PyProtectedMember
    from gevent._ssl3 import SSLSocket


SolBase.voodoo_init()

logger = logging.getLogger(__name__)


class TcpSimpleClient(TcpSocketManager):
    """
    Tcp simple client
    """

    def __init__(self, tcp_client_config):
        """
        Constructor.
        Option via TcpClientConfig to enable SSL on socket.
        In this case, SSL handshake is to be done manually AFTER connect. Higher level implementation MUST handle this.
        :param tcp_client_config: The tcp client config.
        :type tcp_client_config: pysoltcp.tcpclient.TcpClientConfig.TcpClientConfig
        """

        # Base - we provide two callback :
        # - one for disconnecting ourselves
        # - one to notify socket receive buffer
        TcpSocketManager.__init__(self, self.disconnect, self._on_receive)

        # Check
        if tcp_client_config is None:
            logger.error("TcpSimpleClient : tcp_serverConfig is None")
            raise Exception("TcpSimpleClient : tcp_serverConfig is None")
        elif not isinstance(tcp_client_config, TcpClientConfig):
            logger.error("TcpSimpleClient : tcp_serverConfig is not a TcpServerConfig, class=%s", SolBase.get_classname(tcp_client_config))
            raise Exception("TcpSimpleClient : tcp_serverConfig is not a TcpServerConfig")

        # Default
        self._tcp_client_config = tcp_client_config

        # Greenlets
        self._read_greenlet = None
        self._write_greenlet = None

        # Receive queue
        self._receive_queue = Queue()

        # Receive current buffer
        self._receive_current_buf = None

        # Socket info
        self._local_addr = None
        self._local_port = None
        self._remote_addr = None
        self._remote_port = None
        self._connect_count = AtomicIntSafe()
        self._disconnect_count = AtomicIntSafe()

        # Pool stuff
        self.process_id = None
        self.pool_id = None
        self.dt_alloc = SolBase.datecurrent()
        self.dt_last_acquire = self.dt_alloc
        self.dt_last_release = self.dt_alloc
        self.count_acquire = 0
        self.count_release = 0

        # Pool disconnect notify
        self.on_disconnect_notifypoolcallbacklock = Lock()
        self.on_disconnect_notifypoolcallbackcall = True
        self.on_disconnect_notifypoolcallback = None

        # Ssl
        self.ssl_handshake_timeout_ms = 5000

    # ================================
    # POOL METHOD
    # ================================

    def notify_acquire(self):
        """
        Notify
        """
        self.dt_last_acquire = SolBase.datecurrent()
        self.count_acquire += 1

    def notify_release(self):
        """
        Notify
        """
        self.dt_last_release = SolBase.datecurrent()
        self.count_release += 1

    # ================================
    # TO STRING OVERWRITE
    # ================================

    def __str__(self):
        """
        To string override
        :return: A string
        :rtype str
        """

        return "c.addr={0}:{1}/{2}:{3}*cc={4}*dc={5}*c.q.recv.size={6}*sock={7}*proxy={8}/{9}:{10}*{11}".format(
            self._local_addr, self._local_port, self._remote_addr, self._remote_port,
            self._connect_count.get(), self._disconnect_count.get(),
            self._receive_queue.qsize(),
            self.current_socket,
            self._tcp_client_config.proxy_enable,
            self._tcp_client_config.proxy_addr,
            self._tcp_client_config.proxy_port,
            TcpSocketManager.__str__(self)
        )

    # =================================
    # CONNECT
    # =================================

    def connect(self):
        """
        Connect to server. Return true upon success.
        If SSL is enable, do NOT forget at higher level to initiate a SSL handshake manually.
        :return True if connected, false otherwise.
        :rtype bool
        """

        try:
            logger.debug("TcpSimpleClient : connect : starting, target=%s:%s", self._tcp_client_config.target_addr, self._tcp_client_config.target_port)

            # Reset the error flag right now
            self.unset_internal_fatal_error_for_reconnect()

            # Check
            if self.is_connected:
                logger.warning("TcpSimpleClient : connect : already connected, doing nothing, self=%s", self)
                return False

            # Init
            self._connect_count.increment()
            self._dt_created = SolBase.datecurrent()
            self._dt_last_recv = self._dt_created
            self._dt_last_send = self._dt_created

            # ------------------------
            # ALLOC
            # ------------------------
            if self._tcp_client_config.ssl_enable:
                if self._tcp_client_config.proxy_enable:
                    # ------------------------
                    # PROXY ON, SSL ON
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY ON/SSL ON, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY ON/SSL ON, self=%s", self)

                    # Alloc
                    self.current_socket = socks.socksocket()

                    # Proxy
                    self.current_socket.setproxy(socks.PROXY_TYPE_SOCKS5, self._tcp_client_config.proxy_addr, self._tcp_client_config.proxy_port)

                    # SSL Wrap : Inside POST CONNECT
                else:
                    # ------------------------
                    # PROXY OFF, SSL ON
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY OFF/SSL ON, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY OFF/SSL ON, self=%s", self)

                    # Alloc
                    self.current_socket = gevent.socket.socket()

                    # Wrap
                    self.current_socket = SSLSocket(self.current_socket, do_handshake_on_connect=False, ssl_version=PROTOCOL_TLSv1)
            else:
                if self._tcp_client_config.proxy_enable:
                    # ------------------------
                    # PROXY ON, SSL OFF
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY ON/SSL OFF, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY ON/SSL OFF, self=%s", self)

                    # Alloc
                    self.current_socket = socks.socksocket()

                    # Proxy
                    self.current_socket.setproxy(socks.PROXY_TYPE_SOCKS5, self._tcp_client_config.proxy_addr,
                                                 self._tcp_client_config.proxy_port)
                else:
                    # ------------------------
                    # PROXY OFF, SSL OFF
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY OFF/SSL OFF, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY OFF/SSL OFF, self=%s", self)

                    # Alloc
                    self.current_socket = gevent.socket.socket()

            # ------------------------
            # POST ALLOC
            # ------------------------
            if self._tcp_client_config.timeout_ms:
                logger.debug("TcpSimpleClient : connect : setting timeout=%s, self=%s", self._tcp_client_config.timeout_ms, self)
                self.current_socket.settimeout(None)
                logger.debug("TcpSimpleClient : connect : timeout=%s, self=%s", self.current_socket.gettimeout(), self)

            # ------------------------
            # CONNECT
            # ------------------------

            # 2-tuple (host, port)
            address = (self._tcp_client_config.target_addr, self._tcp_client_config.target_port)

            # Connect
            if self._tcp_client_config.debug_log:
                logger.info("TcpSimpleClient : connect : starting, ssl=%s, timeout=%s, self=%s", self._tcp_client_config.ssl_enable, self._tcp_client_config.timeout_ms, self)

            # Go
            logger.debug("TcpSimpleClient : connect : calling connect, self=%s", self)
            self.current_socket.connect(address)
            logger.debug("TcpSimpleClient : connect : now connected, self=%s", self)

            # Store
            self._local_addr = self.current_socket.getsockname()[0]
            self._local_port = self.current_socket.getsockname()[1]
            self._remote_addr = self.current_socket.getpeername()[0]
            self._remote_port = self.current_socket.getpeername()[1]

            # Log
            if self._tcp_client_config.debug_log:
                logger.info("TcpSimpleClient : connected, %s:%s to %s:%s, self=%s", self._local_addr, self._local_port, self._remote_addr, self._remote_port, self)

            # Done
            self.is_connected = True
            self.on_disconnect_notifypoolcallbackcall = True

            # ------------------------
            # POST CONNECT
            # ------------------------
            if self._tcp_client_config.ssl_enable:
                # Switch to SSL now
                if self._tcp_client_config.debug_log:
                    logger.info("PROXY ON/SSL ON, switching to SSL now, self=%s", self)
                else:
                    logger.debug("PROXY ON/SSL ON, switching to SSL now, self=%s", self)

                # Wrap (already connected
                self.current_socket = SSLSocket(self.current_socket, do_handshake_on_connect=False, ssl_version=PROTOCOL_TLSv1)

                # Start
                dt_start = SolBase.datecurrent()

                # Do it
                logger.debug("TcpSimpleClient : __do_ssl_handshake now, self=%s", self)
                self.__do_ssl_handshake()

                # Time
                self._set_ssl_handshake_ms(SolBase.datediff(dt_start))

            # ------------------------
            # POST CONNECT
            # ------------------------

            if self._tcp_client_config.tcp_keepalive_enabled:
                # Switch to TCP KA now
                if self._tcp_client_config.debug_log:
                    logger.info(
                        "TCP KA ON, switching to KA now, (on=%s/delay=%s/failed=%s/interval=%s), self=%s",
                        self._tcp_client_config.tcp_keepalive_enabled,
                        self._tcp_client_config.tcp_keepalive_probes_senddelayms,
                        self._tcp_client_config.tcp_keepalive_probes_failedcount,
                        self._tcp_client_config.tcp_keepalive_probes_sendintervalms,
                        self)
                else:
                    logger.debug(
                        "TCP KA ON, switching to KA now, (on=%s/delay=%s/failed=%s/interval=%s), self=%s",
                        self._tcp_client_config.tcp_keepalive_enabled,
                        self._tcp_client_config.tcp_keepalive_probes_senddelayms,
                        self._tcp_client_config.tcp_keepalive_probes_failedcount,
                        self._tcp_client_config.tcp_keepalive_probes_sendintervalms,
                        self)

                # Go
                self.current_socket.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
                self.current_socket.setsockopt(SOL_TCP, TCP_KEEPIDLE, int(self._tcp_client_config.tcp_keepalive_probes_senddelayms / 1000))
                self.current_socket.setsockopt(SOL_TCP, TCP_KEEPINTVL, int(self._tcp_client_config.tcp_keepalive_probes_sendintervalms / 1000))
                self.current_socket.setsockopt(SOL_TCP, TCP_KEEPCNT, int(self._tcp_client_config.tcp_keepalive_probes_failedcount))

                # Check
                v = self.current_socket.getsockopt(SOL_SOCKET, SO_KEEPALIVE)
                if v != 1:
                    logger.warning("SO_KEEPALIVE mismatch, having=%s, required=1", v)

                v = self.current_socket.getsockopt(SOL_TCP, TCP_KEEPIDLE)
                if v != self._tcp_client_config.tcp_keepalive_probes_senddelayms / 1000:
                    logger.warning("TCP_KEEPIDLE mismatch, having=%s, required=%s", v, self._tcp_client_config.tcp_keepalive_probes_senddelayms / 1000)

                v = self.current_socket.getsockopt(SOL_TCP, TCP_KEEPINTVL)
                if v != self._tcp_client_config.tcp_keepalive_probes_sendintervalms / 1000:
                    logger.warning("TCP_KEEPINTVL mismatch, having=%s, required=%s", v, self._tcp_client_config.tcp_keepalive_probes_sendintervalms / 1000)

                v = self.current_socket.getsockopt(SOL_TCP, TCP_KEEPCNT)
                if v != self._tcp_client_config.tcp_keepalive_probes_failedcount:
                    logger.warning("TCP_KEEPCNT mismatch, having=%s, required=%s", v, self._tcp_client_config.tcp_keepalive_probes_failedcount)

            # Non-blocking mode
            self.current_socket.setblocking(0)
            self.current_socket.settimeout(None)
            logger.debug("TcpSimpleClient : connect : non blocking mode set, self=%s", self)

            # Start the read/write loops
            self._read_greenlet = gevent.spawn(self._read_loop)
            self._write_greenlet = gevent.spawn(self._write_loop)
            logger.debug("TcpSimpleClient : connect : r/w loops started, self=%s", self)

            # Done
            logger.debug("TcpSimpleClient : connect : done, self=%s", self)
            return True

        except Exception as e:
            # Logs
            logger.error("TcpSimpleClient : connect : Exception, ex=%s, self=%s", SolBase.extostr(e), self)

            # Call disconnect
            self._callback_disconnect()

            # Exit
            return False

    # =================================
    # DISCONNECT
    # =================================

    def disconnect(self):
        """
        Disconnect from server. Return true upon success.
        :return True if success, false otherwise.
        :rtype bool
        """
        try:
            logger.debug("TcpSimpleClient : disconnect : entering, self=%s", self)

            # Check
            if not self.is_connected:
                logger.debug("TcpSimpleClient : disconnect : not connected, doing nothing, self=%s", self)
                return False

            # Disconnect
            if self.current_socket:
                logger.debug("TcpSimpleClient : disconnect : socket.shutdown, self=%s", self)
                # noinspection PyUnusedLocal
                try:
                    self.current_socket.shutdown(2)
                except Exception as e:
                    logger.debug("Exception on shutdown, ex=%s, self=%s", SolBase.extostr(e), self)
                    pass

                logger.debug("TcpSimpleClient : disconnect : socket.close, self=%s", self)
                self.current_socket.close()
                self.current_socket = None

            # Reset socket related context
            self._dt_created = SolBase.datecurrent()
            self._dt_last_recv = self._dt_created
            self._dt_last_send = self._dt_created
            self._disconnect_count.increment()
            self._local_addr = None
            self._local_port = None
            self._remote_addr = None
            self._remote_port = None

            # Signal
            logger.debug("TcpSimpleClient : disconnect : is_connected=False, self=%s", self)
            self.is_connected = False

            # Handle notifications at upper level
            go_call = False

            # Second level callback - Process in lock to avoid multiple notification for same simple client instance
            with self.on_disconnect_notifypoolcallbacklock:
                if self.on_disconnect_notifypoolcallbackcall and self.on_disconnect_notifypoolcallback:
                    logger.debug("TcpSimpleClient : disconnect : calling second level disconnect callback, callback=%s, self=%s", self.on_disconnect_notifypoolcallback, self)
                    self.on_disconnect_notifypoolcallbackcall = False
                    go_call = True

            # Out of lock : call
            # (ASYNC, because we may be called by read/write greenlet, which are going to be killed just after)
            if go_call:
                try:
                    # Call it synchronously (mantis 1907)
                    self.on_disconnect_notifypoolcallback(self)

                    # DO NOT MOVE THIS SLEEP, REDIS WILL DEADLOCK ON START
                    SolBase.sleep(0)
                except Exception as e:
                    logger.warning("on_disconnect_notifypoolcallback : call exception=%s", SolBase.extostr(e))

            # Greenlet reset after is_connected=False (will help to exit itself)
            if self._read_greenlet:
                logger.debug("TcpSimpleClient : disconnect : read kill, self=%s", self)
                self._read_greenlet.kill(GreenletExit, False)
                logger.debug("TcpSimpleClient : disconnect : read kill 1, self=%s", self)
                self._read_greenlet = None
                logger.debug("TcpSimpleClient : disconnect : read kill 2, self=%s", self)

            if self._write_greenlet:
                logger.debug("TcpSimpleClient : disconnect : write kill, self=%s", self)
                self._write_greenlet.kill(GreenletExit, False)
                logger.debug("TcpSimpleClient : disconnect : write kill 1, self=%s", self)
                self._write_greenlet = None
                logger.debug("TcpSimpleClient : disconnect : write kill 2, self=%s", self)

            logger.debug("TcpSimpleClient : disconnect : done, self=%s", self)
            return True

        except Exception as e:
            logger.error("TcpSimpleClient : disconnect : Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            return False

    # =================================
    # SSL HANDSHAKE
    # =================================

    def __do_ssl_handshake(self):
        """
        Do a ssl handshake. Method is using a greenlet, but is blocking for the caller.
        """

        # Go
        SolBase.sleep(0)
        try:
            logger.debug("Calling do_handshake now, ssl_handshake_timeout_ms=%s", self.ssl_handshake_timeout_ms)
            with gevent.Timeout(seconds=self.ssl_handshake_timeout_ms / 1000.0, exception=Exception):
                SolBase.sleep(0)
                self.current_socket.do_handshake()
                logger.debug("Called do_handshake")
        except Exception as e:
            logger.warn("Exception while waiting for do_handshake, ex=%s", SolBase.extostr(e))
            self._callback_disconnect()
        finally:
            SolBase.sleep(0)

    def __do_ssl_handshake_internal(self):
        """
        Do a ssl handshake. Call _callback_disconnect() if error.
        """
        try:
            logger.debug("__do_ssl_handshake_internal : start now, self=%s", self)
            SolBase.sleep(0)

        except Exception as e:
            logger.error("__do_ssl_handshake_internal : exception=%s, self=%s", SolBase.extostr(e), self)
            self._callback_disconnect()
        finally:
            SolBase.sleep(0)

    # =================================
    # RECEIVE
    # =================================

    def _on_receive(self, binary_buffer):
        """
        Callback called upon server receive.
        - binary_buffer : a BINARY buffer received on the socket.

        !! CAUTION !!
        - IN ALL CASES THIS METHOD RECEIVE A BINARY BUFFER

        !! CAUTION !!
        - PYTHON 2.X / binary - bytes    => bytes is a byte array (or an ascii string) : nothing to do
        - PYTHON 2.X / text - str  => you receive a bytes, you may convert it to str (SolBase.binary_to_unicode)

        !! CAUTION !!
        - PYTHON 3.X / binary - bytes  => bytes is binary. You may convert it to a bytes using encoding.

                |  2.x                     |  3.x
        --------+--------------------------+-----------------------
        Bytes   |  'abc' <type 'bytes'>      |  b'abc' <type 'bytes'>
        Unicode | u'abc' <type 'str'>  |   'abc' <type 'bytes'>
        :param binary_buffer: Binary buffer received.
        :type binary_buffer: bytes
        """

        # Received something...
        logger.debug("TcpSimpleClient : _on_receive : binary_buffer=%s, self=%s", repr(binary_buffer), self)

        # Parse
        self._receive_current_buf = ProtocolParserTextDelimited.parse_protocol(self._receive_current_buf, binary_buffer, self._receive_queue, b"\n")

    def get_recv_queue_len(self):
        """
        Get receive queue len
        :return The receive queue length.
        :rtype int
        """
        return self._receive_queue.qsize()

    # =================================
    # RECEIVE QUEUE
    # =================================

    def get_from_receive_queue(self, block=False, timeout_sec=None):
        """
        Get a buffer from the receive queue.
        - If block is False, will return an item OR raise an Empty exception if no item.
        - If block is True AND timeOut=None, will wait forever for an item.
        - If block is True and timeout_sec>0, will wait for timeout_sec then raise Empty exception if no item.
        :param block: If true, will block.
        :type block: bool
        :param timeout_sec: The timeout in second.
        :type timeout_sec: None,int
        """

        return self._receive_queue.get(block, timeout_sec)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
import logging
from errno import EWOULDBLOCK
from socket import error
from threading import Lock

import gevent
from gevent import GreenletExit
# noinspection PyProtectedMember
from gevent._socketcommon import wait_read, wait_write
# noinspection PyProtectedMember
from gevent.queue import Queue, Empty
from pysolbase.SolBase import SolBase
from pysolmeters.Meters import Meters

from pysoltcp import string_types, binary_type
from pysoltcp.tcpbase.SignaledBuffer import SignaledBuffer

SolBase.voodoo_init()
logger = logging.getLogger(__name__)


class TcpSocketManager(object):
    """
    Tcp socket manager.
    """

    # Send queue : Wait in ms per loop.
    # Upon queue fill, called is unlocked immediately.
    # We use a call go get(Block=True, QUEUE_WAIT_MS_PER_LOOP) to avoid eating cpu while waiting.
    # Value in second OR None for unlimited wait.
    QUEUE_WAIT_SEC_PER_LOOP = None

    def __init__(self, callback_disconnect, callback_receive):
        """
        Constructor.
        :param callback_disconnect: Callback to call upon socket disconnection.
        :type callback_disconnect: Callable
        :param callback_receive:  Callback to call upon socket receive.
        :type callback_receive: Callable
        """

        # Connected flag (backed by is_connected property)
        self.is_connected = False

        # Socket instance
        self._socket = None

        # Send queue
        self.__send_queue = Queue()

        # Callback : called when something is received from the socket
        self._callback_receive = callback_receive

        # Callback : called upon socket disconnection
        self._callback_disconnect = callback_disconnect

        # Timestamps
        self._dt_created = SolBase.datecurrent()
        self._dt_last_recv = self._dt_created
        self._dt_last_send = self._dt_created

        # SSL Handshake asynch
        self.__ssl_handshake_asynch = False
        self.__ssl_handshake_pending = False
        self.__ssl_handshake_timeout_ms = None
        self.__ssl_wait_debug_ms = None
        self.__ssl_timeout_greenlet = None
        self.__ssl_locker = Lock()

        # SSL handshake time
        self.__ssl_handshake_ms = None

        # Fatal error
        self.__internal_fatal_error = False

    # ================================
    # TO STRING OVERWRITE
    # ================================

    def __str__(self):
        """
        To string override
        :return: A string
        :rtype str
        """

        to = None
        if self.current_socket:
            to = self.current_socket.timeout
        return "s.to={0}*dt.creat={1}*dt.recv={2}*dt.send={3}*q.send.size={4}*is.co={5}/{6}*cl={7}*id={8}".format(
            to,
            int(SolBase.datediff(self._dt_created) * 0.001),
            int(SolBase.datediff(self._dt_last_recv) * 0.001),
            int(SolBase.datediff(self._dt_last_send) * 0.001),
            self.__send_queue.qsize(),
            self.is_connected,
            self.__is_running(),
            self.__class__.__name__,
            id(self)
        )

    def set_callback_receive(self, callback_receive):
        """
        Set the receive callback
        :param callback_receive: Callback
        :type callback_receive: Callable
        """

        self._callback_receive = callback_receive

    # ================================
    # SSL HANDSHAKE TIMEOUT
    # ================================

    def __shedule_ssl_handshake_timeout(self):
        """
        Shedule a ssl handshake timeout
        """

        with self.__ssl_locker:
            logger.debug("__shedule_ssl_handshake_timeout : ms=%s, self=%s", self.__ssl_handshake_timeout_ms, self)
            self.__ssl_timeout_greenlet = gevent.spawn_later(self.__ssl_handshake_timeout_ms * 0.001, self.__on_ssl_timeout)

    def _unschedule_ssl_handshake_timeout(self):
        """
        Unschedule the ssl timeout
        """
        with self.__ssl_locker:
            logger.debug("_unschedule_ssl_handshake_timeout, self=%s", self)

            if self.__ssl_timeout_greenlet:
                self.__ssl_timeout_greenlet.kill()
                self.__ssl_timeout_greenlet = None

    def __on_ssl_timeout(self):
        """
        Called when SSL timeout.
        """

        # Reset first (in LOCK)
        with self.__ssl_locker:
            self.__ssl_timeout_greenlet = None

        # Process (outside of LOCK, high level may call _unschedule_ssl_handshake_timeout in lock)
        logger.debug("__on_ssl_timeout, self=%s", self)

        # Check
        if not self.__ssl_handshake_pending:
            # Done, exit
            return
        elif not self.__is_running():
            # Not connected, exit
            return

        # Not done, FATAL
        logger.warning("handshake timeout, fatal, ms=%s, self=%s", self.__ssl_handshake_timeout_ms, self)

        # Stat
        Meters.aii("tcp.server.ssl_handshake_timeout_count")

        # Timeout, Fatal, exit now
        self._disconnect_helper("__on_ssl_timeout")

    # ================================
    # GETTER/SETTER
    # ================================

    def set_ssl_handshake_asynch(self, ssl_handshake_asynch_enable, ssl_handshake_timeout_ms, ssl_wait_debug_ms=None):
        """
        Enable or disable ssl asynch handshake.
        :param ssl_handshake_asynch_enable: If True, will enable it.
        :type ssl_handshake_asynch_enable: bool
        :param ssl_handshake_timeout_ms: Apply if asynch handshake is enable. Timeout in ms for handshake itself.
        :type ssl_handshake_timeout_ms: int
        :param ssl_wait_debug_ms: Debug only.
        :type ssl_wait_debug_ms: int,None
        """

        self.__ssl_handshake_asynch = ssl_handshake_asynch_enable
        self.__ssl_handshake_timeout_ms = ssl_handshake_timeout_ms
        self.__ssl_wait_debug_ms = ssl_wait_debug_ms

        # If enable, switch to pending true
        if self.__ssl_handshake_asynch:
            # Now pending
            self.__ssl_handshake_pending = True

            # Schedule a greenlet, waiting for ssl to complete
            self.__shedule_ssl_handshake_timeout()

    def _set_ssl_handshake_ms(self, ms):
        """
        Set the time taken to perform the ssl handshake.
        :param ms: Millis.
        :type ms: int        
        """
        self.__ssl_handshake_ms = ms

    def _set_socket(self, my_soc):
        """
        Setter
        :param my_soc: The socket.
        :type my_soc: socket.socket
        """
        self._socket = my_soc

    def _set_is_connected(self, b):
        """
        Setter
        :param b: New value.
        :type b: bool
        """
        self.__isConnected = b

    def _get_ssl_handshake_ms(self):
        """
        Return the ssl handshake elapsed ms.
        :return: Integer or None if not set.
        :rtype int,None
        """
        return self.__ssl_handshake_ms

    def _get_socket(self):
        """
        Getter.
        :return: The current socket.
        :rtype socket.socket
        """
        return self._socket

    def _get_is_connected(self):
        """
        Getter.
        :return: A boolean.
        :rtype bool
        """
        return self.__isConnected

    def _get_send_queue(self):
        """
        Getter.
        :return: The gevent send queue.
        :rtype gevent.queue.Queue
        """
        return self.__send_queue

    def _get_dt_created(self):
        """
        Getter.
        :return: A datetime.
        :rtype datetime
        """
        return self._dt_created

    def _get_dt_last_send(self):
        """
        Getter.
        :return: A datetime.
        :rtype datetime
        """
        return self._dt_last_send

    def _get_dt_last_recv(self):
        """
        Getter.
        :return: A datetime.
        :rtype datetime
        """
        return self._dt_last_recv

    # ================================
    # PROPERTIES
    # ================================

    # Properties
    is_connected = property(_get_is_connected, _set_is_connected)
    current_socket = property(_get_socket, _set_socket)
    send_queue = property(_get_send_queue)

    # Properties : timestamp
    date_created = property(_get_dt_created)
    date_last_send = property(_get_dt_last_send)
    date_last_recv = property(_get_dt_last_recv)

    # ================================
    # SEND TO SOCKET
    # ================================

    def send_binary_to_socket(self, buffer_to_send):
        """
        Send to socket, asynch.
        :param buffer_to_send: The localBuffer to send (bytes)
        :type buffer_to_send: bytes
        :return: True is send has been scheduled, false otherwise.
        :rtype bool
        """

        # Check
        if not isinstance(buffer_to_send, binary_type):
            logger.error("buffer_to_send not a binary, class=%s, self=%s", SolBase.get_classname(buffer_to_send), self)
            return False

        # Check
        if not self.__is_running():
            logger.debug("not connected, returning false, self=%s", self)
            return False

        # Enqueue
        self.__send_queue.put(buffer_to_send)

        # Stats
        Meters.aii("tcp.server.server_bytes_send_pending", len(buffer_to_send))

        return True

    def send_binary_to_socket_with_signal(self, signaled_buffer):
        """
        Send to socket, asynch.
        Upon completion, signaled_buffer.send_event is set.
        Caution : Caller MUST check the boolean returned. If False, the event will NOT be set.
        :param signaled_buffer: A signaled_buffer instance.
        :type signaled_buffer: pysoltcp.tcpbase.SignaledBuffer.SignaledBuffer
        :return: True is send has been scheduled, false otherwise.
        :rtype bool
        """

        # Check
        if not isinstance(signaled_buffer, SignaledBuffer):
            logger.error("signaled_buffer not a SignaledBuffer, class=%s, self=%s", SolBase.get_classname(signaled_buffer), self)
            return False

            # Check
        if not isinstance(signaled_buffer.binary_buffer, binary_type):
            logger.error("binary_buffer not a binary, class=%s, self=%s", SolBase.get_classname(signaled_buffer.binary_buffer), self)
            return False

        # Check
        if not self.__is_running():
            logger.debug(
                "not connected, returning false, self=%s",
                self)
            return False

        # Enqueue
        self.__send_queue.put(signaled_buffer)

        # Stats
        Meters.aii("tcp.server.server_bytes_send_pending", len(signaled_buffer.binary_buffer))

        return True

    def send_text_to_socket(self, text_to_send, append_lf=True):
        """
        Send text to socket, asynch.
        :param text_to_send: The text to send (str)
        :type text_to_send: str
        :param append_lf: If true, append an \n
        :return: True is send has been scheduled, false otherwise.
        """

        # DEPRECATED, use send_unicode_to_socket
        return self.send_unicode_to_socket(text_to_send, append_lf=append_lf)

    def send_unicode_to_socket(self, unicode_to_send, encoding="utf-8", append_lf=True):
        """
        Send text to socket, asynch.
        :param unicode_to_send: The text to send (str)
        :type unicode_to_send: str
        :param encoding: The encoding to use.
        :type encoding: str
        :param append_lf: If true, append an \n
        :type append_lf: bool
        :return: True is send has been scheduled, false otherwise.
        :rtype: bool
        """

        # Check
        if not isinstance(unicode_to_send, string_types):
            logger.error("unicode_to_send not an string_types, class=%s, str=%s, self=%s", SolBase.get_classname(unicode_to_send), repr(unicode_to_send), self)
            return False

        # Go
        unicode_temp = unicode_to_send

        # LF if required
        if append_lf:
            unicode_temp += u"\n"

        # Convert to binary localBuffer
        bin_buf = SolBase.unicode_to_binary(unicode_temp, encoding)

        # Send binary
        return self.send_binary_to_socket(bin_buf)

    def get_send_queue_len(self):
        """
        Return the send queue len.
        :return The queue length (integer)
        :rtype int
        """
        return self.__send_queue.qsize()

    # ================================
    # DISCONNECT HELPER
    # ================================

    def unset_internal_fatal_error_for_reconnect(self):
        """"
        This must be called before firing a reconnect AFTER a disconnection
        """

        self.__internal_fatal_error = False

    def _disconnect_helper(self, reason):
        """
        Fire a disconnect on our end and notify upper level is possible
        :param reason: Reason
        :type reason: str
        """

        logger.debug("Disconnect from inside, reason=%s", reason)

        # Disengage r/w loops now
        # CAUTION : NEVER (NEVER) internally flag "is_connected=False" here, you will BLAST upper level implementations
        self.__internal_fatal_error = True

        # Close the socket
        SolBase.safe_close_socket(self.current_socket)
        self.current_socket = None

        # Callback now
        if self._callback_disconnect:
            self._callback_disconnect()
        else:
            logger.error("_disconnect_helper is None, check your implementation, self=%s", self)

    def __is_running(self):
        """
        Return true if we are connected and we do not have a fatal error
        :return: bool
        :rtype bool
        """

        return self.is_connected and not self.__internal_fatal_error

    # ================================
    # SOCKET WAIT
    # ================================

    def _wait_for_socket_recv(self):
        """
        Wait for socket available for read (recv).
        :return Return True if available, False otherwise.
        :rtype bool
        """

        # Check
        if not self.__is_running():
            logger.debug("not connected, returning false, self=%s", self)
            return False

        # Go
        try:
            # Wait
            wait_read(self._socket.fileno())

            # Ready
            return True
        except Exception as e:
            logger.warning("Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            return False

    def _wait_for_socket_send(self):
        """
        Wait for socket available for write (send).
        :return Return True if available, False otherwise.
        :rtype bool
        """

        # Check
        if not self.__is_running():
            logger.debug("not connected, returning false, self=%s", self)
            return False

        # Go
        try:
            # Wait
            wait_write(self.current_socket.fileno())

            # Ready
            return True
        except Exception as e:
            logger.warning("Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            return False

    # ================================
    # SOCKET LOOP : READ
    # ================================

    def _read_loop(self):
        """
        High level read loop on socket
        """
        logger.debug("entering now, self=%s", self)
        try:
            self._read_loop_internal()
        except GreenletExit:
            logger.debug("exiting due to GreenletExit, self=%s", self)
            return
        except Exception as e:
            logger.error("Exception raised, ex=%s, self=%s", SolBase.extostr(e), self)
        finally:
            logger.debug("exiting now, self=%s", self)
            SolBase.sleep(0)

    def _read_loop_internal(self):
        """
        Low level read loop on socket
        """
        logger.debug("entering now, self=%s", self)
        try:
            while self.__is_running():
                try:
                    if self.__ssl_handshake_pending:
                        # Pending SSL handshake + received something
                        # Handle SSL handshake now

                        # Stats
                        Meters.dtci("tcp.server.delay_server_accept_to_sslhandshakestart", SolBase.datediff(self._dt_created))

                        # Timestamps
                        self._dt_last_recv = SolBase.datecurrent()

                        # Debug ONLY
                        if self.__ssl_wait_debug_ms:
                            logger.warning("DEBUG : forcing a wait for SSL handshake timeout, ms=%s, self=%s", self.__ssl_wait_debug_ms, self)
                            SolBase.sleep(self.__ssl_wait_debug_ms)
                            logger.warning("DEBUG : forcing a wait for SSL handshake timeout, done, self=%s", self)

                        # Do the handshake
                        self.current_socket.do_handshake()

                        # Done, cancel timeout
                        self._unschedule_ssl_handshake_timeout()

                        # Ms
                        ms = SolBase.datediff(self._dt_last_recv)

                        # SSL Stats (for client)
                        self._set_ssl_handshake_ms(ms)

                        # Server stats
                        Meters.dtci("tcp.server.delay_server_sslhandshake", ms)

                        # Done
                        self.__ssl_handshake_pending = False

                        # Non blocking mode now
                        self.current_socket.setblocking(0)
                        self.current_socket.settimeout(None)

                        # Reloop in normal mode
                        continue

                    # Wait for socket to be available for read
                    ok = self._wait_for_socket_recv()
                    if not ok:
                        # This is not really normal
                        logger.warning("_wait_for_socket_recv returned False, self=%s", self)
                    elif not self.__is_running():
                        logger.debug("_wait_for_socket_recv returned True, __is_running()==False, exiting, self=%s", self)
                        return
                    else:
                        # Something to read...
                        local_buf = self._read_from_socket()
                        if not self.__is_running():
                            logger.debug("_read_from_socket returned, __is_running()==False, exiting, self=%s", self)
                        elif local_buf is None:
                            # This is not really normal
                            logger.debug("_read_from_socket returned None, self=%s", self)
                        elif len(local_buf) == 0:
                            # This is not really normal
                            logger.debug("_read_from_socket returned empty string, self=%s", self)
                            # Gevent 1.0.2 : call disconnect
                            self._disconnect_helper("_read_from_socket returned empty string")
                        else:
                            # Timestamps
                            self._dt_last_recv = SolBase.datecurrent()

                            # Notify
                            if self._callback_receive:
                                self._callback_receive(local_buf)
                            else:
                                logger.error("_callback_receive is None, check you implementation, self=%s", self)

                    # Next read
                    SolBase.sleep(0)
                except Exception as e:
                    logger.warning("IN_LOOP Exception raised, ex=%s, self=%s", SolBase.extostr(e), self)
        except Exception as e:
            logger.error("METHOD Exception raised, ex=%s, self=%s", SolBase.extostr(e), self)
        finally:
            logger.debug("exiting now, self=%s", self)
            SolBase.sleep(0)

    def _read_from_socket(self):
        """
        Read from the socket. Return a local_buf or None.
        """

        # Check
        if not self.__is_running():
            logger.debug("not connected, doing nothing, self=%s", self)
            return None

        try:
            # TODO : Socket local_buf size
            # TODO : Socket, upon server closure, will BLOCK on recv. Upon time, auto disconnect the client.

            # Try to read
            local_buf = self.current_socket.recv(1024)

            # If local_buf is empty string, socket is disconnected.
            if local_buf and len(local_buf) == 0:
                # Disconnect
                logger.debug("empty string received, disconnecting ourselves, self=%s", self)
                self._disconnect_helper("_read_from_socket : Empty string")
                return None

            # Stats
            Meters.aii("tcp.server.server_bytes_received", len(local_buf))

            # Ok
            return local_buf
        except error as e:
            # [Errno 11] Resource temporarily unavailable
            # Means that nothing is available to read.
            # If not this, we raise.
            if e.args[0] != EWOULDBLOCK:
                # Raise :)
                logger.debug("_tryReadWrite : _read_from_socket : error, ex=%s, self=%s", SolBase.extostr(e), self)
                self._disconnect_helper("_read_from_socket : error / No EWOULDBLOCK")
                return None
            else:
                # Normal
                logger.debug("_tryReadWrite : _read_from_socket : normal exception/EWOULDBLOCK, ex=%s, self=%s", e,
                             self)
                return None
        except Exception as e:
            logger.debug("_tryReadWrite : _read_from_socket : Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            self._disconnect_helper("_read_from_socket : Exception / No EWOULDBLOCK")
            return None

    # ================================
    # SOCKET LOOP : WRITE
    # ================================

    def _write_loop(self):
        """
        High level read loop on socket
        """
        logger.debug("entering now, self=%s", self)
        try:
            self._write_loop_internal()
        except GreenletExit:
            logger.debug("exiting due to GreenletExit, self=%s", self)
            return
        except Exception as e:
            logger.error("Exception raised, ex=%s, self=%s", SolBase.extostr(e), self)
        finally:
            logger.debug("exiting now, , self=%s", self)
            SolBase.sleep(0)

    def _write_loop_internal(self):
        """
        Low level read/write loop on socket
        """
        logger.debug("entering now, self=%s", self)
        try:
            while self.__is_running():
                try:
                    # Wait for the queue
                    try:
                        # Call, with blocking.
                        item = self.send_queue.get(True, TcpSocketManager.QUEUE_WAIT_SEC_PER_LOOP)

                    except Empty:
                        # Next read
                        SolBase.sleep(0)
                        continue

                    # Go some
                    if isinstance(item, binary_type):
                        # Length
                        length = len(item)
                        # Buffer to send
                        buf_to_send = item
                        # Event to signal
                        event_to_signal = None
                    elif isinstance(item, SignaledBuffer):
                        # Stats
                        length = len(item.binary_buffer)
                        # Buffer to send
                        buf_to_send = item.binary_buffer
                        # Event to signal
                        event_to_signal = item.send_event
                    else:
                        logger.warning("not managed class in queue, class=%s, item=%s", SolBase.get_classname(item), item)
                        continue

                    # Stat now
                    Meters.aii("tcp.server.server_bytes_send_pending", -length)

                    # Wait for socket to be available for write
                    logger.debug("waiting for socket to send=%s, self=%s", repr(item), self)
                    ok = self._wait_for_socket_send()
                    if not ok:
                        # This is not really normal
                        if self.__is_running():
                            logger.warning("_wait_for_socket_send returned False (running true), self=%s", self)
                        else:
                            logger.info("_wait_for_socket_send returned False (running false), self=%s", self)
                    elif not self.__is_running():
                        logger.debug("Ready for send, but __is_running==False, exiting, send=%s, self=%s", self)
                        return
                    else:
                        try:
                            # Ready to write, fire
                            logger.debug("writing to socket to send=%s, self=%s", repr(item), self)
                            ok = self._write_to_socket(buf_to_send)
                            if not ok:
                                if self.__is_running():
                                    # This is not really normal
                                    logger.warning("_write_to_socket returned False (running true), self=%s", self)
                                else:
                                    logger.warning("_write_to_socket returned False (running false), self=%s", self)

                            # Signal if applicable
                            if event_to_signal:
                                event_to_signal.set()
                        finally:
                            # Timestamps
                            self._dt_last_send = SolBase.datecurrent()

                            # Stats
                            Meters.aii("tcp.server.server_bytes_send_done", length)

                    # Next read
                    SolBase.sleep(0)
                except Exception as e:
                    logger.warning("IN_LOOP Exception raised, ex=%s, self=%s", SolBase.extostr(e), self)
        except Exception as e:
            logger.error("METHOD Exception raised, ex=%s, self=%s", SolBase.extostr(e), self)
        finally:
            logger.debug("exiting now, self=%s", self)
            SolBase.sleep(0)

    def _write_to_socket(self, local_buffer):
        """
        Write to the socket.
        """

        # Check
        if not self.__is_running():
            logger.debug("not connected, doing nothing, self=%s", self)
            return False

        try:
            #
            self.current_socket.sendall(local_buffer)
            return True
        except error as e:
            # [Errno 11] Resource temporarily unavailable
            # Means that nothing is available to read.
            # If not this, we raise.
            if e.args[0] != EWOULDBLOCK:
                # Raise :)
                logger.debug("Exception, ex=%s, self=%s", SolBase.extostr(e), self)
                self._disconnect_helper("_write_to_socket : error / No EWOULDBLOCK")
                return False
            else:
                # Normal
                logger.debug("normal exception/EWOULDBLOCK, ex=%s, self=%s", e, self)
                return False
        except Exception as e:
            logger.info("Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            self._disconnect_helper("_write_to_socket : Exception")
            return False



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
import logging
import sys

logger = logging.getLogger(__name__)

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] >= 3

# Switches:
# PY2        PY3
# str        bytes
# unicode    str

# Code :
# a) replace str by bytes
# b) replace unicode by str

# Then :
if PY3:
    string_types = str,
    integer_types = int,
    text_type = str
    binary_type = bytes,bytearray

    items = dict.items
    itervalues = dict.values
    # noinspection PyShadowingBuiltins
    xrange = range
    # noinspection PyShadowingBuiltins
    long = int

    max_int = sys.maxsize

    print("str={0}".format(str))
    print("bytes={0}".format(bytes))
else:
    # noinspection PyUnresolvedReferences,PyCompatibility
    string_types = basestring,
    # noinspection PyUnresolvedReferences,PyUnboundLocalVariable
    integer_types = (int, long)
    # noinspection PyUnresolvedReferences
    text_type = unicode, str
    binary_type = str

    # noinspection PyUnresolvedReferences
    items = dict.items
    # noinspection PyUnresolvedReferences
    itervalues = dict.itervalues
    # noinspection PyUnresolvedReferences
    import __builtin__

    # noinspection PyUnresolvedReferences,PyShadowingBuiltins
    xrange = __builtin__.xrange

    # noinspection PyShadowingBuiltins
    # noinspection PyUnresolvedReferences
    str = unicode

    # noinspection PyUnresolvedReferences
    max_int = sys.maxint

    # noinspection PyUnresolvedReferences
    print("unicode={0}".format(unicode))
    print("str={0}".format(str))
    print("bytes={0}".format(bytes))



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
# Logger
import logging

from pysolbase.SolBase import SolBase

SolBase.logging_init()
logger = logging.getLogger(__name__)


class ProtocolParserTextDelimited(object):
    """
    A protocol parser, able to react to chunk of protocols received onto a socket.
    """

    @classmethod
    def parse_protocol(cls, previous_buffer, input_buffer, gevent_queue, separator=b"\n", single_item_processing=False):
        """
        Parse a received protocol.
        - For all COMPLETE token : enqueue in the provided queue (method : put)
        - Return the remaining INCOMPLETE token (ie not \n terminated)
        - Return None if no INCOMPLETE token remains
        :param cls: Our class.
        :param previous_buffer: The previous incomplete buffer.
        :type previous_buffer: bytes,None
        :param input_buffer: The new buffer.
        :type input_buffer: bytes,None
        :param gevent_queue: A gevent queue, which will be populated with complete items.
        :type gevent_queue: Queue
        :param separator: The separator to use
        :type separator: bytes
        :param single_item_processing: If true, process only one item then exit.
        :type single_item_processing: bool
        :return: The new incomplete buffer.
        :rtype bytes
        """

        # - ar[0] : a string
        # - ar[1] : IF separator found : separator
        # - ar[1] : IF separator not found : empty
        # - ar[2] : remaining part or empty

        # Split
        prev_buf = previous_buffer
        if input_buffer:
            cur_buf = input_buffer
        else:
            cur_buf = b""

        # Prefix by prev_buf
        if prev_buf:
            cur_buf = prev_buf + cur_buf
            prev_buf = None

        while True:
            # Need separator as bytes
            # noinspection PyTypeChecker
            ar = cur_buf.partition(separator)

            # Check ar[1]
            if len(ar[1]) > 0:
                # Separator found, having a COMPLETE token in ar[0] : Enqueue it, with optional previous buffer prefix
                if prev_buf is None:
                    gevent_queue.put(ar[0])
                else:
                    gevent_queue.put(previous_buffer + ar[0])
                    # noinspection PyUnusedLocal
                    prev_buf = None

                # Check if we have remaining stuff
                if len(ar[2]) == 0:
                    # Nothing more
                    return prev_buf
                else:
                    # Remaining stuff : continue to parse
                    if not single_item_processing:
                        # noinspection PyUnusedLocal
                        cur_buf = ar[2]
                    else:
                        return ar[2]
            else:
                # Separator not found
                # We have an INCOMPLETE token
                # Stop parse and return it
                # BUT only IF not empty
                if len(ar[0]) == 0:
                    if prev_buf is None:
                        return None
                    elif len(prev_buf) > 0:
                        return prev_buf
                    else:
                        return None

                else:
                    if prev_buf is None:
                        return ar[0]
                    else:
                        return prev_buf + ar[0]



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

import logging
import signal
import traceback
from _ssl import PROTOCOL_TLSv1
from threading import RLock, Lock

import gevent
import os
from gevent.event import Event
from gevent.server import StreamServer
from gevent.timeout import Timeout
from pysolbase.SolBase import SolBase
from pysolmeters.AtomicInt import AtomicIntSafe
from pysolmeters.Meters import Meters

from pysoltcp.tcpbase.TcpSocketManager import TcpSocketManager
from pysoltcp.tcpserver.TcpServerConfig import TcpServerConfig

SolBase.voodoo_init()
logger = logging.getLogger(__name__)


# noinspection PyAbstractClass
class StreamServerNoClose(StreamServer):
    """
    Hack
    """

    def do_close(self, *args):
        """
        Do nothing
        """

        pass


class TcpServer(object):
    """
    Fast tcp server, gevent (epoll) based.
    Option via TcpServerConfig to enable SSL.
    If enabled, SSL handshake mode is done AFTER accept (do_handshake_on_connect=False)
    """

    def __init__(self, tcp_server_config):
        """
        Constructor.
        :param tcp_server_config: The configuration.
        :type tcp_server_config: pysoltcp.tcpserver.TcpServerConfig.TcpServerConfig
        """

        # Check
        if tcp_server_config is None:
            logger.error("tcp_server_config is None")
            raise Exception("tcp_server_config is None")
        elif not isinstance(tcp_server_config, TcpServerConfig):
            logger.error("tcp_server_config is not a TcpServerConfig, class=%s", SolBase.get_classname(tcp_server_config))
            raise Exception("tcp_server_config is not a TcpServerConfig")

        # Store
        self._tcp_server_config = tcp_server_config

        # Init =>

        # _is_started :
        # - True if server is started, False is server is stopped.
        # - Start : try to start, if ok set to True
        # - Stop : stop server, stop client, and set to False
        # - Means : During stop ongoing, will be True
        self._is_started = False

        # _is_running :
        # - True if server is running, False is server is no more running
        # - Start : Set to true, try to start, if failed, set to False
        # - Stop : Set to False, stop server, stop client
        # - Means : During stop ongoing, will be False
        self._is_running = False

        # Gevent StreamServer
        self._server = None

        # Fork pids
        self._fork_pid_list = list()

        # Client management (re-entrant lock)
        self._client_connected_atomicint = AtomicIntSafe()
        self._client_connected_hash = dict()
        self._client_connected_hash_lock = RLock()

        # Lock for start/stop
        self.__stop_start_lock = Lock()

        # Control variables
        self._effective_control_interval_ms = 0

        # Control init
        self.__set_effective_controlinterval_ms()

        # Auto start
        if self._tcp_server_config.auto_start is True:
            logger.info("Auto-starting ON, starting now")
            self.start_server()
        else:
            logger.info("Auto-starting OFF")

    # =====================================================
    # START / STOP : HIGH LEVEL
    # =====================================================

    def start_server(self):
        """
        Start

        """

        with self.__stop_start_lock:
            logger.info("start_server : starting")
            try:
                if self._is_started is True:
                    logger.warning("start_server : already started, doing nothing")
                    return False

                # Running
                self._is_running = True

                # Low level start
                self._start_server()

                # Start
                self._is_started = True

                # Done
                logger.info("start_server : started, %s", SolBase.get_current_pid_as_string())

                # Exit
                return True
            except Exception as e:
                logger.error("start_server : exception, ex=%s", SolBase.extostr(e))
                # Failed, not more running
                self._is_running = False

                # Raise
                raise

    def destroy(self):
        """
        For spring python. Just call stop_server().
        """
        logger.info("destroy : Entering, calling self.stop_server(), %s", SolBase.get_current_pid_as_string())
        self.stop_server()

    def stop_server(self):
        """
        Stop server
        """

        with self.__stop_start_lock:
            logger.info("stop_server : stopping, %s", SolBase.get_current_pid_as_string())
            try:
                # No more running
                self._is_running = False

                # Check
                if self._is_started is False:
                    logger.info("stop_server : already stopped, doing nothing, %s", SolBase.get_current_pid_as_string())
                    return

                # Low level stop
                try:
                    gevent.with_timeout(self._tcp_server_config.stop_server_timeout_ms, self._stop_server)
                except Timeout:
                    logger.warning("Timeout while calling low level _stop_server, some socket may be stucked")

                # Stop
                self._is_started = False

                # Done
                logger.info("stop_server : stopped, %s", SolBase.get_current_pid_as_string())
            except Exception as e:
                logger.error("stop_server : exception, ex=%s, %s", SolBase.extostr(e), SolBase.get_current_pid_as_string())
                raise

    # =====================================================
    # CONFIG ACCESS
    # =====================================================

    def get_tcpserver_config(self):
        """
        Return the configuration associated to the tcpserver.
        :return: A TcpServerConfig instance.
        :rtype pysoltcp.tcpserver.TcpServerConfig.TcpServerConfig
        """

        return self._tcp_server_config

    # =====================================================
    # START / STOP : LOW LEVEL
    # =====================================================

    def _start_server(self):
        """
        Low level start
        """

        # Allocate a server, and provide a connection callback
        logger.info("listen_addr=%s", self._tcp_server_config.listen_addr)
        logger.info("listen_port=%s", self._tcp_server_config.listen_port)
        logger.info("ssl_enable=%s", self._tcp_server_config.ssl_enable)
        logger.info("ssl_key_file=%s", self._tcp_server_config.ssl_key_file)
        logger.info("ssl_certificate_file=%s", self._tcp_server_config.ssl_certificate_file)

        logger.info("child_process_count=%s", self._tcp_server_config.child_process_count)

        logger.info("ssl_handshake_timeout_ms=%s", self._tcp_server_config.ssl_handshake_timeout_ms)

        logger.info("onstop_call_client_stopsynch=%s", self._tcp_server_config.onstop_call_client_stopsynch)

        logger.info("socket_absolute_timeout_ms=%s", self._tcp_server_config.socket_absolute_timeout_ms)
        logger.info("socket_relative_timeout_ms=%s", self._tcp_server_config.socket_relative_timeout_ms)
        logger.info("socket_min_checkinterval_ms=%s", self._tcp_server_config.socket_min_checkinterval_ms)

        logger.info("_effective_control_interval_ms=%s", self._effective_control_interval_ms)

        logger.info("stop_client_timeout_ms=%s", self._tcp_server_config.stop_client_timeout_ms)
        logger.info("stop_server_timeout_ms=%s", self._tcp_server_config.stop_server_timeout_ms)

        logger.info("client_factory=%s", self._tcp_server_config.client_factory)

        if self._tcp_server_config.ssl_enable is False:
            # No SSL
            logger.info("Starting in TCP/CLEAR mode")
            self._server = StreamServerNoClose(
                (self._tcp_server_config.listen_addr,
                 self._tcp_server_config.listen_port),
                self._on_connection)
        else:
            # SSL ON
            logger.info("Starting in TCP/SSL mode")
            self._server = StreamServerNoClose(
                (self._tcp_server_config.listen_addr,
                 self._tcp_server_config.listen_port),
                self._on_connection,
                # SSL enabling
                keyfile=self._tcp_server_config.ssl_key_file, certfile=self._tcp_server_config.ssl_certificate_file,
                # SSL handshake after accept
                do_handshake_on_connect=True,
                # TLS
                ssl_version=PROTOCOL_TLSv1,
                # Cipher
                # ciphers="RC4-MD5",
                # Args
            )

            self._server.min_delay = 0.0
            self._server.max_delay = 0.0

        # Startup
        if self._tcp_server_config.child_process_count <= 0:
            # Normal start-up
            logger.info("Starting in NON-FORKED mode")
            self._server.start()
        else:
            # Child process startup : prestart
            logger.info("Pre-starting in FORKED mode, subprocess=%s", self._tcp_server_config.child_process_count)
            # GEVENT_RC1 fix : pre_start => init_socket
            self._server.init_socket()

            # Let's rock
            logger.info("Forking gevent")
            for idx in range(self._tcp_server_config.child_process_count):
                # Fork gevent hub
                # GEVENT_RC1 fix : hub.fork => os.fork
                _fork_pid = gevent.fork()

                logger.info("Forking gevent hub done, idx=%s, forkPid=%s, %s", idx, _fork_pid, SolBase.get_current_pid_as_string())

                # Check it
                if _fork_pid == 0:
                    # We are in a child => exit this loop
                    SolBase.set_master_process(False)
                    break
                else:
                    # Master on
                    SolBase.set_master_process(True)

                    # Store child pid
                    logger.info("Storing child _fork_pid=%s", _fork_pid)
                    self._fork_pid_list.append(int(_fork_pid))

            # Start accepting now (parent and all sub-processes)
            logger.info("Accepting now, %s", SolBase.get_current_pid_as_string())
            self._server.start_accepting()

    def _stop_server(self):
        """
        Low level stop
        """

        try:
            # If we have child, signal them now
            for pid in self._fork_pid_list:
                logger.info("_stop_server : Sending SIGTERM to pid=%s", pid)
                os.kill(pid, signal.SIGTERM)

            # Wait for exit
            for pid in self._fork_pid_list:
                logger.info("_stop_server : Waiting for pid=%s", pid)

                # Wait
                wait_pid, wait_status = os.waitpid(pid, 0)

                # Get result
                wait_signal = wait_status & 0xff
                if wait_signal == 0:
                    wait_code = wait_status > 8
                else:
                    wait_code = 0

                # Info
                logger.info("_stop_server : Waiting for pid=%s ok, wait_status=%s, wait_signal=%s, wait_code=%s, wait_pid=%s", pid, wait_status, wait_signal, wait_code, wait_pid)

            # Stop (timeout = 5 seconds)
            self._server.stop(5)

            # Clear client
            self._remove_all_client()
        finally:
            # Reset
            self._server = None

    # =====================================================
    # INTERNAL CALLBACKS
    # =====================================================

    def _on_connection(self, socket, address):
        """
        Callback called upon client connection.
        :param socket:  The client socket.
        :type socket: socket.socket
        :param address: The client remote address:
        :type address: str

        """
        logger.debug("_on_connection : address=%s %s", address, SolBase.get_current_pid_as_string())

        # Register a new session
        # This will start the read/write loop on client.
        local_client = self._register_client(socket, address)

        # Check
        if local_client is None:
            logger.error("_on_connection : _register_client returned none")

    # =====================================================
    # CONTROL INTERVAL HELPER
    # =====================================================

    def __set_effective_controlinterval_ms(self):
        """
        Set the effective control interval in ms
        """

        # Get values
        val_abs = self._tcp_server_config.socket_absolute_timeout_ms
        val_rel = self._tcp_server_config.socket_relative_timeout_ms

        # If absolute and relative both lower then zero : nothing to do, socket has no limit
        if val_abs <= 0 and val_rel <= 0:
            self._effective_control_interval_ms = 0
        else:
            # Here, one of those is greater than zero.
            # If all of them are greater than zero : we keep the minimum
            if val_abs > 0 and val_rel > 0:
                val_sch = min(val_abs, val_rel)
            # Else, one of them if zero or lower : we keep the maximum
            else:
                val_sch = max(val_abs, val_rel)

            # We have val_sch, which is our target control check interval in ms.
            # To avoid too low values (and too high check frequency), we use correct it with minimal check interval
            self._effective_control_interval_ms = max(val_sch, self._tcp_server_config.socket_min_checkinterval_ms)

    def get_effective_controlinterval_ms(self):
        """
        Return the effective control interval in ms to apply to socket
        :return: An integer (millis)
        :rtype int
        """

        return self._effective_control_interval_ms

    # =====================================================
    # CLIENT MANAGEMENT : REGISTER
    # =====================================================

    def _register_client(self, socket, address):
        """
        Register a new client.
        :param socket:  The client socket.
        :type socket: socket.socket
        :param address: The client remote address
        :type address: str
        :return Return a TcpServerClientContext upon success, None upon failure.
        :rtype pysoltcp.tcpserver.clientcontext.TcpServerClientContext.TcpServerClientContext
        """

        try:
            logger.debug("entering")

            # Must be started
            if self._is_started is False:
                logger.debug("not started, cannot process")
                return None

            # Allocate a new client context
            logger.debug("allocating new_client using factory")
            new_client = self._tcp_server_config.client_factory.get_new_clientcontext(
                self,
                self._client_connected_atomicint.increment(),
                socket, address)

            # Hash id
            logger.debug("hashing new_client")
            with self._client_connected_hash_lock:
                logger.debug("hashing new_client (in lock)")
                self._client_connected_hash[new_client.get_client_id()] = new_client

            # Statistics
            logger.debug("populating statistics")
            Meters.aii("tcp.server.client_connected")
            Meters.aii("tcp.server.client_register_count")

            # Enable SSL if required and set handshake timeout
            if self._tcp_server_config.ssl_enable is True:
                new_client.set_ssl_handshake_asynch(True, self._tcp_server_config.ssl_handshake_timeout_ms, self._tcp_server_config.debug_waitinsslms)
            # Start the client
            logger.debug("starting client")
            new_client.start()

            # Log
            logger.debug("client started and hashed, id=%s, addr=%s", new_client.get_client_id(), new_client.get_client_addr())
            return new_client
        except Exception as e:
            # Error
            logger.warning("extostr, ex=%s", SolBase.extostr(e))

            # Statistics
            Meters.aii("tcp.server.clientRegisterException")

            # Close the socket in this case
            SolBase.safe_close_socket(socket)
            return None

    # =====================================================
    # CLIENT MANAGEMENT : REMOVE (ASYNCH and SYNCH) and REMOVE ALL
    # =====================================================

    def _remove_client_asynch(self, client_id):
        """
        Remove a client, asynch.
        :param client_id: The client id.
        :type client_id: int
        """

        # Spawn
        logger.debug("entering, client_id=%s", client_id)

        # Signal event (mantis 1280)
        evt = Event()

        # Spawn
        gevent.spawn(self._remove_client, client_id, evt)

        # Switch
        SolBase.sleep(0)

        # And wait
        # Note : remove this wait do not impact unittest...
        logger.debug("waiting, client_id=%s", client_id)
        evt.wait()

        # Over
        logger.debug("done, client_id=%s", client_id)

    # noinspection PyMethodMayBeStatic
    def _remove_client_stop_internal(self, old_client, evt):
        """
        Remove internal
        :param old_client: oldclient
        :type: old_client: pysoltcp.tcpserver.clientcontext.TcpServerClientContext.TcpServerClientContext
        :param evt: gevent.Event
        :type evt: gevent.Event
        """

        try:
            # Get
            cid = old_client.get_client_id()

            # Stop the client r/w loops and close the sock
            logger.debug("_remove_client_stop_internal call, cid=%s", cid)
            old_client.stop_synch_internal()
        except Exception as e:
            logger.warning("Ex=%s", SolBase.extostr(e))
        finally:
            evt.set()

    def _remove_client_stop_business(self, old_client, evt):
        """
        Remove internal
        :param old_client: oldclient
        :type old_client: pysoltcp.tcpserver.clientcontext.TcpServerClientContext.TcpServerClientContext
        :param evt: gevent.Event
        :type evt: gevent.Event
        """

        # -------------------------
        # Stop the client (Business call here)
        # We do NOT call if :
        # - service is stopping AND onstop_call_client_stopsynch==False
        # -------------------------

        try:
            cid = old_client.get_client_id()

            logger.debug("_remove_client_stop_business call, cid=%s", cid)

            if self._is_running:
                # -------------------
                # Running, call
                # -------------------
                logger.debug("stop_synch call (_is_running==%s), cid=%s", self._is_running, cid)
                old_client.stop_synch()
            elif not self._is_running and self._tcp_server_config.onstop_call_client_stopsynch:
                # -------------------
                # Not running + call ON : call
                # -------------------
                logger.debug("stop_synch call (_is_running==%s + onstop_call_client_stopsynch==%s), cid=%s", self._is_running, self._tcp_server_config.onstop_call_client_stopsynch, cid)
                old_client.stop_synch()
            else:
                # -------------------
                # No call
                # -------------------
                logger.debug("stop_synch NOT CALLED (_is_running==%s + onstop_call_client_stopsynch==%s), cid=%s", self._is_running, self._tcp_server_config.onstop_call_client_stopsynch, cid)
                pass
        except Exception as e:
            logger.warning("Ex=%s", SolBase.extostr(e))
        finally:
            evt.set()

    def _remove_client(self, client_id, evt):
        """
        Remove a client. Return a TcpServerClientContext upon success,
        :param client_id: The client id.
        :type client_id: int
        :param evt: Event to signal
        :type evt: gevent.Event, None
        :return The removed TcpServerClientContext or None upon failure.
        :rtype None,pysoltcp.tcpserver.clientcontext.TcpServerClientContext.TcpServerClientContext
        """

        logger.debug("entering, client_id=%s", client_id)

        try:
            with self._client_connected_hash_lock:
                # Check
                if client_id not in self._client_connected_hash:
                    # Note : This may occurs in some conditions
                    logger.debug("client_id not hashed, id=%s", client_id)
                    Meters.aii("tcp.server.client_remove_nothashed")
                    return None

                # Get (direct, we are already in lock)
                old_client = self._get_client_fromid(client_id)

                # Remove from hashmap
                logger.debug("un-hashing, client_id=%s", client_id)
                del (self._client_connected_hash[client_id])

            # ------------------------------
            # Out of lock : call async : BUSINESS
            # ------------------------------
            try:
                local_evt = Event()
                g1 = gevent.spawn(self._remove_client_stop_business, old_client, local_evt)
                SolBase.sleep(0)

                local_evt.wait(self._tcp_server_config.stop_client_timeout_ms / 1000.0)

                if not local_evt.isSet():
                    # Flush out warning
                    s = "Greenlet dump, g={0}, frame={1}".format(g1, ''.join(traceback.format_stack(g1.gr_frame)))

                    # Cleanup
                    s = s.replace("\n", " # ")
                    while s.find("  ") >= 0:
                        s = s.replace("  ", " ")

                    # Error logs
                    logger.error("Timeout in _remove_client_stop_business client_id=%s, stack=%s", client_id, s)

                    # Kill
                    g1.kill(block=True)

                    # Stat
                    Meters.aii("tcp.server.client_remove_timeout_business")

            except Exception as e:
                logger.warning("Exception in _remove_client_stop_business client_id=%s, ex=%s", client_id, SolBase.extostr(e))

            # ------------------------------
            # Out of lock : call async : INTERNAL
            # ------------------------------
            try:
                local_evt = Event()
                g2 = gevent.spawn(self._remove_client_stop_internal, old_client, local_evt)
                SolBase.sleep(0)

                local_evt.wait(self._tcp_server_config.stop_client_timeout_ms / 1000.0)

                if not local_evt.isSet():
                    # Flush out warning
                    s = "Greenlet dump, g={0}, frame={1}".format(g2, ''.join(traceback.format_stack(g2.gr_frame)))

                    # Cleanup
                    s = s.replace("\n", " # ")
                    while s.find("  ") >= 0:
                        s = s.replace("  ", " ")

                    # Error logs
                    logger.error("Timeout in _remove_client_stop_internal client_id=%s, stack=%s", client_id, s)

                    # Kill
                    g2.kill(block=True)

                    # Stat
                    Meters.aii("tcp.server.client_remove_timeout_internal")
            except Exception as e:
                logger.warning("Exception in _remove_client_stop_internal client_id=%s, ex=%s", client_id, SolBase.extostr(e))

            # Statistics
            Meters.aii("tcp.server.client_connected", -1)
            Meters.aii("tcp.server.client_remove_count")

            # Log
            logger.debug("client removed, id=%s, addr=%s", old_client.get_client_id(), old_client.get_client_addr())

            return old_client
        except Exception as e:
            # Error
            logger.warning("Exception, ex=%s", SolBase.extostr(e))

            # Statistics
            Meters.aii("tcp.server.client_remove_exception")

            return None
        finally:
            if evt:
                evt.set()

    def _remove_all_client(self):
        """
        Remove all clients.
        """
        try:
            # PY3 : Cannot iterate through keys (it do not copy anymore)
            # => we force a list alloc

            # Pass through all client and remove all
            for client_id in list(self._client_connected_hash.keys()):
                self._remove_client(client_id, None)
        except Exception as e:
            # Error
            logger.warning("_remove_all_client : Exception, ex=%s", SolBase.extostr(e))

    # =====================================================
    # CLIENT MANAGEMENT : GET FROM HASHMAP
    # =====================================================

    def _get_client_fromid(self, client_id):
        """
        Get a client from id. Return None if not found.
        :param client_id: The client Id
        :type client_id: int
        :return A TcpServerClientContext or None if not found.
        :rtype None,pysoltcp.tcpserver.clientcontext.TcpServerClientContext.TcpServerClientContext
        """
        with self._client_connected_hash_lock:
            # Check
            if client_id not in self._client_connected_hash:
                return None
            else:
                return self._client_connected_hash[client_id]



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

import logging

from pysolbase.FileUtility import FileUtility
from pysolbase.SolBase import SolBase

from pysoltcp.tcpserver.basefactory.TcpServerClientContextAbstractFactory import TcpServerClientContextAbstractFactory
from pysoltcp.tcpserver.clientcontext.TcpServerClientContextFactory import TcpServerClientContextFactory

SolBase.logging_init()
logger = logging.getLogger(__name__)


class TcpServerConfig(object):
    """
    tcpserver configuration.
    If ssl is NOT enable, it is better that calls to sslCertificate/ssl_key_file should not happen.
    """

    def __init__(self):
        """
        Constructor.
        """

        logger.debug("TcpServerConfig : Constructor called")

        # Default
        self._auto_start = False
        self._listen_addr = None
        self._listen_port = None
        self._client_factory = None
        self._set_client_factory(TcpServerClientContextFactory())

        # Ssl
        self._ssl_enable = False
        self._ssl_key_file = None
        self._ssl_certificate_file = None
        self._ssl_handshake_timeout_ms = 10000

        # Stop behavior
        self._onstop_callclient_stopsynch = False

        # Forking
        self._child_process_count = 0

        # Debug
        self._debug_waitinssl_ms = None

        # Socket control : Absolute default = 7 days (x 24 hours, x60 min, x60 sec, x1000 for ms)
        self._socket_absolute_timeout_ms = 7 * 24 * 60 * 60 * 1000

        # Socket control : Relative default = 1 days (x 24 hours, x60 min, x60 sec, x1000 for ms)
        self._socket_relative_timeout_ms = 1 * 24 * 60 * 60 * 1000

        # Socket control : minimal interval between 2 check = 60 sec
        self._socket_min_check_interval_ms = 60000

        # Timeout while stopping client
        self._stop_client_timeout_ms = 10000

        # Timeout while stopping server
        self._stop_server_timeout_ms = 30000

    def _set_client_factory(self, client_factory):
        """
        Setter. Raise exception if a problem occurs.
        :param client_factory : must be an instance of the factory, not the class itself.
        :type client_factory: pysoltcp.tcpserver.basefactory.TcpServerClientContextAbstractFactory.TcpServerClientContextAbstractFactory

        """
        if client_factory is None:
            logger.error("client_factory none")
            raise Exception("client_factory none")
        elif not isinstance(client_factory, TcpServerClientContextAbstractFactory):
            logger.error("client_factory is not a TcpServerClientContextAbstractFactory")
            raise Exception("client_factory is not a TcpServerClientContextAbstractFactory")
        else:
            self._client_factory = client_factory

    def _set_listen_addr(self, listen_addr):
        """
        Setter. Raise exception if a problem occurs.
        :param listen_addr: The listen address.
        :type listen_addr: str
        """

        self._listen_addr = listen_addr

    def _set_listen_port(self, listen_port):
        """
        Setter. Raise exception if a problem occurs.
        :param listen_port: The listen port.
        :type listen_port: int
        """

        listen_port = SolBase.to_int(listen_port)

        if not SolBase.is_int(listen_port):
            logger.error("not a int, class=%s", SolBase.get_classname(listen_port))
            raise Exception("not a int")
        elif listen_port == 0:
            logger.warning("newPort==0")
            raise Exception("newPort==0")
        else:
            self._listen_port = listen_port

    def _set_ssl_handshake_timeout_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: Time out in ms.
        :type ms: int
        """
        cms = SolBase.to_int(ms)

        if not SolBase.is_int(cms):
            logger.error("not a int, class=%s", SolBase.get_classname(cms))
            raise Exception("not a int")
        else:
            self._ssl_handshake_timeout_ms = cms

    def _set_stop_client_timeout_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: Time out in ms.
        :type ms: int
        """

        cms = SolBase.to_int(ms)

        if not SolBase.is_int(cms):
            logger.error("not a int, class=%s", SolBase.get_classname(cms))
            raise Exception("not a int")
        else:
            self._stop_client_timeout_ms = cms

    def _set_stop_server_timeout_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: Time out in ms.
        :type ms: int
        """

        cms = SolBase.to_int(ms)
        if not SolBase.is_int(cms):
            logger.error("not a int, class=%s", SolBase.get_classname(cms))
            raise Exception("not a int")
        else:
            self._stop_server_timeout_ms = cms

    def _set_child_process_count(self, cpc):
        """
        Setter. Raise exception if a problem occurs.
        :param cpc: The number of child process count. If zero, no fork is performed (default).
        :type cpc: int
        """

        child_process_count = SolBase.to_int(cpc)

        if not SolBase.is_int(child_process_count):
            logger.error("not a int, class=%s", SolBase.get_classname(child_process_count))
            raise Exception("not a int")
        elif child_process_count < 0:
            logger.warning("child_process_count<0")
            raise Exception("child_process_count<0")
        else:
            self._child_process_count = child_process_count

    def _set_auto_start(self, b):
        """
        Getter
        :param b: A boolean.
        :type b: bool
        """

        mb = SolBase.to_bool(b)

        if not SolBase.is_bool(mb):
            logger.error("not a boolean, class=%s", SolBase.get_classname(mb))
            raise Exception("not a boolean")
        else:
            self._auto_start = mb

    def _set_onstop_call_client_stopsynch(self, b):
        """
        Getter
        :param b: A boolean.
        :type b: bool
        """

        mb = SolBase.to_bool(b)

        if not SolBase.is_bool(mb):
            logger.error("not a boolean, class=%s", SolBase.get_classname(mb))
            raise Exception("not a boolean")
        else:
            self._onstop_callclient_stopsynch = mb

    def _set_ssl_enable(self, b):
        """
        Enable or disable ssl.
        :param b: Boolean.
        :type b: bool
        """

        is_enable = SolBase.to_bool(b)

        if not SolBase.is_bool(is_enable):
            logger.error("not a boolean, class=%s", SolBase.get_classname(is_enable))
            raise Exception("not a boolean")
        else:
            self._ssl_enable = is_enable

    def _set_ssl_keyfile(self, key_file):
        """
        Set the ssl key file. Raise exception upon error.
        :param key_file: A valid key file. Must exist and be accessible.
        :type key_file: str
        :return: True upon success, false otherwise.
        :rtype bool
        """
        # Check
        if key_file is None or len(key_file) == 0:
            self._ssl_key_file = ""
            return

        if not FileUtility.is_file_exist(key_file):
            logger.warning("key_file do not exist or is not accessible, key_file=%s", key_file)

        # Set
        self._ssl_key_file = key_file

    def _set_ssl_certificate_file(self, cert_file):
        """
        Set the ssl certificate file. Raise exception upon error.
        :param cert_file: A valid certificate file. Must exist and be accessible.
        :type cert_file: str
        """

        # Check
        if cert_file is None or len(cert_file) == 0:
            self._ssl_certificate_file = ""
            return

        # Check
        if not FileUtility.is_file_exist(cert_file):
            logger.warning("cert_file do not exist or is not accessible, cert_file=%s", cert_file)

        # Set
        self._ssl_certificate_file = cert_file

    def _get_client_factory(self):
        """
        Getter
        :return The client factory.
        :rtype pysoltcp.tcpserver.basefactory.TcpServerClientContextAbstractFactory.TcpServerClientContextAbstractFactory
        """
        return self._client_factory

    def _get_listen_port(self):
        """
        Getter
        :return The listen port.
        :rtype int
        """
        return self._listen_port

    def _get_ssl_handshake_timeout_ms(self):
        """
        Getter
        :return The value.
        :rtype int
        """
        return self._ssl_handshake_timeout_ms

    def _get_stop_client_timeout_ms(self):
        """
        Getter
        :return The value.
        :rtype int
        """
        return self._stop_client_timeout_ms

    def _get_stop_server_timeout_ms(self):
        """
        Getter
        :return The value.
        :rtype int
        """
        return self._stop_server_timeout_ms

    def _get_child_process_count(self):
        """
        Getter
        :return The number of child process count.
        :rtype int
        """
        return self._child_process_count

    def _get_auto_start(self):
        """
        Getter
        :return The auto start boolean
        :rtype bool
        """
        return self._auto_start

    def _get_onstop_call_client_stopsynch(self):
        """
        Getter
        :return bool
        :rtype bool
        """
        return self._onstop_callclient_stopsynch

    def _get_listen_addr(self):
        """
        Getter
        :return The listen address.
        :rtype str
        """
        return self._listen_addr

    def _get_ssl_enable(self):
        """
        Getter.
        :return: A boolean.
        :rtype bool
        """
        return self._ssl_enable

    def _get_ssl_keyfile(self):
        """
        Getter.
        :return: A string or None
        :rtype None,str
        """

        return self._ssl_key_file

    def _get_ssl_certificate_file(self):
        """
        Getter.
        :return: A string or None
        :rtype None,str
        """

        return self._ssl_certificate_file

    def _get_socket_absolute_timeout_ms(self):
        """
        Getter
        :return Timeout in ms.
        :rtype int
        """
        return self._socket_absolute_timeout_ms

    def _get_socket_relative_timeout_ms(self):
        """
        Getter
        :return Timeout in ms.
        :rtype int
        """
        return self._socket_relative_timeout_ms

    def _get_socket_min_checkinterval_ms(self):
        """
        Getter
        :return Timeout in ms.
        :rtype int
        """
        return self._socket_min_check_interval_ms

    def _set_socket_min_checkinterval_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: Timeout in ms.
        :type ms: int
        """

        cms = SolBase.to_int(ms)

        if not SolBase.is_int(cms):
            logger.error("not a int, class=%s", SolBase.get_classname(cms))
            raise Exception("not a int")
        else:
            self._socket_min_check_interval_ms = cms

    def _set_socket_absolute_timeout_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: Timeout in ms.
        :type ms: int
        """

        cms = SolBase.to_int(ms)

        if not SolBase.is_int(cms):
            logger.error("not a int, class=%s", SolBase.get_classname(cms))
            raise Exception("not a int")
        else:
            self._socket_absolute_timeout_ms = cms

    def _set_socket_relative_timeout_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: Timeout in ms.
        :type ms: int
        """

        cms = SolBase.to_int(ms)

        if not SolBase.is_int(cms):
            logger.error("not a int, class=%s", SolBase.get_classname(cms))
            raise Exception("not a int")
        else:
            self._socket_relative_timeout_ms = cms

    # ========================
    # Properties - configs
    # ========================

    # Auto-start on ?
    auto_start = property(_get_auto_start, _set_auto_start)
    # Listen address
    listen_addr = property(_get_listen_addr, _set_listen_addr)
    # Listen port
    listen_port = property(_get_listen_port, _set_listen_port)
    # Client factory
    client_factory = property(_get_client_factory, _set_client_factory)
    # Ssl on ?
    ssl_enable = property(_get_ssl_enable, _set_ssl_enable)
    # Ssl on ?
    onstop_call_client_stopsynch = property(_get_onstop_call_client_stopsynch, _set_onstop_call_client_stopsynch)
    # Ssl key file
    ssl_key_file = property(_get_ssl_keyfile, _set_ssl_keyfile)
    # Ssl certificate file
    ssl_certificate_file = property(_get_ssl_certificate_file, _set_ssl_certificate_file)

    # ========================
    # Properties - tuning
    # ========================

    # Child process count
    child_process_count = property(_get_child_process_count, _set_child_process_count)
    # Timeout in ms allowed for SSL handshake
    ssl_handshake_timeout_ms = property(_get_ssl_handshake_timeout_ms, _set_ssl_handshake_timeout_ms)
    # Per socket : absolute socket time out in ms. If lt or eq to 0 : no limit.
    socket_absolute_timeout_ms = property(_get_socket_absolute_timeout_ms, _set_socket_absolute_timeout_ms)
    # Per socket : relative socket time out in ms. If lt or eq to 0 : no limit.
    socket_relative_timeout_ms = property(_get_socket_relative_timeout_ms, _set_socket_relative_timeout_ms)
    # Socket : Minimum check interval in ms.
    socket_min_checkinterval_ms = property(_get_socket_min_checkinterval_ms, _set_socket_min_checkinterval_ms)
    # Stop timeout ms
    stop_client_timeout_ms = property(_get_stop_client_timeout_ms, _set_stop_client_timeout_ms)
    # Stop timeout ms
    stop_server_timeout_ms = property(_get_stop_server_timeout_ms, _set_stop_server_timeout_ms)

    # ========================
    # Properties - DEBUG ONLY - NEVER USE THEM IN PRODUCTION
    # ========================

    def _set_debug_waitinsslms(self, value):
        """
        Setter for debug only.
        :param value: Value.
        :type value: int
        """
        self._debug_waitinssl_ms = value

    def _get_debug_waitinsslms(self):
        """
        Getter for debug only.
        :return: Value.
        :rtype int
        """
        return self._debug_waitinssl_ms

    debug_waitinsslms = property(_get_debug_waitinsslms, _set_debug_waitinsslms)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# Logger
import logging

from pysolbase.SolBase import SolBase

SolBase.logging_init()
logger = logging.getLogger(__name__)


class TcpClientConfig(object):
    """
    Tcp client configuration.
    """

    def __init__(self):
        """
        Constructor.
        """

        # Default
        self._target_addr = None
        self._target_port = None

        # Ssl
        self._ssl_enable = False

        # Timeout
        self._timeout_ms = 10000

        # Debug logs
        self._debug_log = False

        # Socks5 proxy
        self._proxy_enable = False
        self._proxy_addr = None
        self._proxy_port = None

        # Tcp keep alive on by default
        self._tcp_keepalive_enabled = True
        self._tcp_keepalive_probes_senddelayms = 60000
        self._tcp_keepalive_probes_failedcount = 5
        self._tcp_keepalive_probes_sendintervalms = 60000

    def _set_targetaddr(self, target_addr):
        """
        Setter. Raise exception if a problem occurs.
        :param target_addr: The target address.
        :type target_addr: str
        """

        self._target_addr = target_addr

    def _get_targetaddr(self):
        """
        Getter
        :return: The target address.
        :rtype str
        """
        return self._target_addr

    def _set_target_port(self, target_port):
        """
        Setter. Raise exception if a problem occurs.
        :param target_port: The target port.
        :type target_port: int
        """
        if not SolBase.is_int(target_port):
            logger.error("TcpClientConfig : _set_target_port : not a int, class=%s", SolBase.get_classname(target_port))
            raise Exception("TcpClientConfig : _set_target_port : not a int")
        elif target_port == 0:
            logger.warning("TcpClientConfig : _set_target_port : newPort==0")
            raise Exception("TcpClientConfig : _set_target_port : newPort==0")
        else:
            self._target_port = target_port

    def _set_timeout_ms(self, ms):
        """
        Setter. Raise exception if a problem occurs.
        :param ms: The timeout in millis or None.
        :type ms: int,None
        """
        self._timeout_ms = ms

    def _get_target_port(self):
        """
        Getter
        :return The target port.
        :rtype int
        """
        return self._target_port

    def _set_ssl_enable(self, b):
        """
        Enable or disable ssl.
        :param b: bool.
        :type b: bool
        """
        self._ssl_enable = b

    def _set_debug_log(self, b):
        """
        Enable or disable debug log.
        :param b: bool.
        :type b: bool
        """
        self._debug_log = b

    def _get_ssl_enable(self):
        """
        Getter.
        :return: A boolean.
        :rtype bool
        """
        return self._ssl_enable

    def _get_timeout_ms(self):
        """
        Getter.
        :return: An integer or None.
        :rtype int,None
        """
        return self._timeout_ms

    def _get_debug_log(self):
        """
        Getter.
        :return: An integer or None.
        :rtype int,None
        """
        return self._debug_log

    def _set_proxy_addr(self, proxy_addr):
        """
        Setter. Raise exception if a problem occurs.
        :param proxy_addr: The proxy target address.
        :type proxy_addr: str
        """

        self._proxy_addr = proxy_addr

    def _get_proxy_addr(self):
        """
        Getter
        :return: The proxy target address.
        :rtype str
        """
        return self._proxy_addr

    def _set_proxy_port(self, proxy_port):
        """
        Setter. Raise exception if a problem occurs.
        :param proxy_port: The target proxy port.
        :type proxy_port: int
        """
        if not SolBase.is_int(proxy_port):
            logger.error("TcpClientConfig : _set_proxy_port : not a int, class=%s", SolBase.get_classname(proxy_port))
            raise Exception("TcpClientConfig : _set_proxy_port : not a int")
        elif proxy_port == 0:
            logger.warning("TcpClientConfig : _set_proxy_port : newPort==0")
            raise Exception("TcpClientConfig : _set_proxy_port : newPort==0")
        else:
            self._proxy_port = proxy_port

    def _get_proxy_port(self):
        """
        Getter
        :return The target port.
        :rtype int
        """
        return self._proxy_port

    def _set_proxy_enable(self, is_enable):
        """
        Enable or disable proxy.
        :param is_enable: Boolean.
        :type is_enable: bool
        """
        self._proxy_enable = is_enable

    def _get_proxy_enable(self):
        """
        Getter.
        :return: A boolean.
        :rtype bool
        """
        return self._proxy_enable

    # =======================
    # TCP KEEP ALIVE
    # =======================

    def _set_tcp_keepalive_enabled(self, is_enable):
        """
        Enable or disable tcp keep alive.
        :param is_enable: Boolean.
        :type is_enable: bool
        :return: Nothing.
        """
        self._tcp_keepalive_enabled = is_enable

    def _get_tcp_keepalive_enabled(self):
        """
        Getter.
        :return: A boolean.
        :rtype bool
        """
        return self._tcp_keepalive_enabled

    def _set_tcp_keepalive_probes_senddelayms(self, value):
        """
        Setter. Raise exception if a problem occurs.
        :param value: Value.
        :type value: int
        
        """
        if not SolBase.is_int(value):
            logger.error("TcpClientConfig : _set_tcp_keepalive_probes_senddelayms : not a int, class=%s", SolBase.get_classname(value))
            raise Exception("TcpClientConfig : _set_tcp_keepalive_probes_senddelayms : not a int")
        elif value == 0:
            logger.warning("TcpClientConfig : _set_tcp_keepalive_probes_senddelayms : newPort==0")
            raise Exception("TcpClientConfig : _set_tcp_keepalive_probes_senddelayms : newPort==0")
        else:
            self._tcp_keepalive_probes_senddelayms = value

    def _get_tcp_keepalive_probes_senddelayms(self):
        """
        Getter
        :return Value
        :rtype int
        """
        return self._tcp_keepalive_probes_senddelayms

    def _set_tcpkeepalive_probes_failedcount(self, value):
        """
        Setter. Raise exception if a problem occurs.
        :param value: Value.
        :type value: int
        
        """
        if not SolBase.is_int(value):
            logger.error("TcpClientConfig : _set_tcpkeepalive_probes_failedcount : not a int, class=%s", SolBase.get_classname(value))
            raise Exception("TcpClientConfig : _set_tcpkeepalive_probes_failedcount : not a int")
        elif value == 0:
            logger.warning("TcpClientConfig : _set_tcpkeepalive_probes_failedcount : newPort==0")
            raise Exception("TcpClientConfig : _set_tcpkeepalive_probes_failedcount : newPort==0")
        else:
            self._tcp_keepalive_probes_failedcount = value

    def _get_tcpkeepalive_probes_failedcount(self):
        """
        Getter
        :return Value
        :rtype int
        """
        return self._tcp_keepalive_probes_failedcount

    def _set_tcpkeepalive_probes_sendintervalms(self, value):
        """
        Setter. Raise exception if a problem occurs.
        :param value: Value.
        :type value: int
        
        """
        if not SolBase.is_int(value):
            logger.error("TcpClientConfig : _set_tcpkeepalive_probes_sendintervalms : not a int, class=%s", SolBase.get_classname(value))
            raise Exception("TcpClientConfig : _set_tcpkeepalive_probes_sendintervalms : not a int")
        elif value == 0:
            logger.warning("TcpClientConfig : _set_tcpkeepalive_probes_sendintervalms : newPort==0")
            raise Exception("TcpClientConfig : _set_tcpkeepalive_probes_sendintervalms : newPort==0")
        else:
            self._tcp_keepalive_probes_sendintervalms = value

    def _get_tcpkeepalive_probes_sendintervalms(self):
        """
        Getter
        :return Value
        :rtype int
        """
        return self._tcp_keepalive_probes_sendintervalms

    # Properties
    target_addr = property(_get_targetaddr, _set_targetaddr)
    target_port = property(_get_target_port, _set_target_port)
    ssl_enable = property(_get_ssl_enable, _set_ssl_enable)
    timeout_ms = property(_get_timeout_ms, _set_timeout_ms)
    debug_log = property(_get_debug_log, _set_debug_log)

    proxy_enable = property(_get_proxy_enable, _set_proxy_enable)
    proxy_addr = property(_get_proxy_addr, _set_proxy_addr)
    proxy_port = property(_get_proxy_port, _set_proxy_port)

    tcp_keepalive_enabled = property(_get_tcp_keepalive_enabled, _set_tcp_keepalive_enabled)
    tcp_keepalive_probes_senddelayms = property(_get_tcp_keepalive_probes_senddelayms, _set_tcp_keepalive_probes_senddelayms)
    tcp_keepalive_probes_failedcount = property(_get_tcpkeepalive_probes_failedcount, _set_tcpkeepalive_probes_failedcount)
    tcp_keepalive_probes_sendintervalms = property(_get_tcpkeepalive_probes_sendintervalms, _set_tcpkeepalive_probes_sendintervalms)
